/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/vfs.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/mount.h>

#include "vfs.h"
#include "misc.h"
#include "reiserfs_fs.h"
#include "reiserfs_fs_sb.h"
#include "reiserfs.h"
#include "resize.h"

static unsigned long int_node_cnt   = 0, int_moved_cnt   = 0;
static unsigned long leaf_node_cnt  = 0, leaf_moved_cnt  = 0;
static unsigned long unfm_node_cnt  = 0, unfm_moved_cnt  = 0;
static __u32 total_node_cnt = 0;
static unsigned long total_moved_cnt = 0;

static unsigned long unused_block;
static unsigned long blocks_used;
static int block_count_mismatch = 0;
static struct bitmap_head * bmp;

/* abnornal exit from block reallocation process */
static void quit_resizer()
{
	/* save changes to bitmap blocks */
	sync_bitmap (bmp);
	free_bitmap (bmp);
	/* leave fs in ERROR state */
	die ("resize_reiserfs: fs shrinking was not completed successfully, run reiserfsck.\n");
}

/* block moving */
static unsigned long move_generic_block(struct super_block *s, unsigned long block, unsigned long bnd, int h)
{
    struct buffer_head * bh, * bh2;

	/* primitive fsck */
	if (block > SB_BLOCK_COUNT(s)) {
		fprintf(stderr, "resize_reiserfs: invalid block number (%lu) found.\n", block);
		quit_resizer();
	}
	/* progress bar, 3D style :) */
	if (opt_verbose)
	    print_how_far(&total_node_cnt, (__u32)blocks_used);
	else
		total_node_cnt ++;

	/* infinite loop check */
	if( total_node_cnt > blocks_used && !block_count_mismatch) {
		fputs("resize_reiserfs: warning: block count exeeded\n",stderr);
		block_count_mismatch = 1;
	}

	if (block < bnd) /* block will not be moved */
		return 0;
	
	/* move wrong block */ 
	bh = bread(s->s_dev, block, s->s_blocksize);

	unused_block = find_1st_unused_block_right(bmp, unused_block);
	if (unused_block == 0 || unused_block >= bnd) {
		fputs ("resize_reiserfs: can\'t find free block\n", stderr);
		quit_resizer();
	}

	/* blocknr changing */
	bh2 = getblk(s->s_dev, unused_block, s->s_blocksize);
	memcpy(bh2->b_data, bh->b_data, bh2->b_size);
	mark_block_free(bmp, block);
	mark_block_used(bmp, unused_block);

	brelse(bh);
	mark_buffer_uptodate(bh2,1);
	mark_buffer_dirty(bh2,0);
	bwrite(bh2);
	brelse(bh2);

	total_moved_cnt++;
	return unused_block;
}

static unsigned long move_unformatted_block(struct super_block *s, unsigned long block, unsigned long bnd, int h)
{
	unsigned long b;
	unfm_node_cnt++;
	b = move_generic_block(s, block, bnd, h);
	if (b)
		unfm_moved_cnt++;
	return b;		
}


/* recursive function processing all tree nodes */
static unsigned long move_formatted_block(struct super_block *s, unsigned long block, unsigned long bnd, int h)
{
    struct buffer_head * bh;
    struct item_head *ih;
    unsigned long new_blocknr = 0;
    int node_is_internal = 0;
    int i, j;
	
    bh = bread(s->s_dev, block, s->s_blocksize);

    if (B_IS_ITEMS_LEVEL(bh)) { /* leaf node*/

	leaf_node_cnt++;

	for (i=0; i < B_NR_ITEMS(bh); i++) {
	    ih = B_N_PITEM_HEAD(bh, i);
	    if (is_indirect_le_ih(ih)) {
		for (j = 0; j < I_UNFM_NUM(ih); j++) {
		    unsigned long  unfm_block;
		    if (B_I_POS_UNFM_POINTER(bh, ih, j) == 0) /* hole */
			continue;
		    unfm_block = move_unformatted_block(s, B_I_POS_UNFM_POINTER(bh, ih, j), bnd, h + 1);
		    if (unfm_block) {
			PUT_B_I_POS_UNFM_POINTER(bh, ih, j, unfm_block);
			mark_buffer_dirty(bh,0);
		    }
		}
	    }	
	}
    } else if (B_IS_KEYS_LEVEL(bh)) { /* internal node */

	int_node_cnt++;
	node_is_internal = 1;

	for (i=0; i <= B_NR_ITEMS(bh); i++) {
	    unsigned long moved_block;
	    moved_block = move_formatted_block(s, B_N_CHILD_NUM(bh, i), bnd, h+1);
	    if (moved_block) {
		PUT_B_N_CHILD_NUM(bh, i, moved_block);
		mark_buffer_dirty(bh,0);
	    }
	}	
    } else {
	die ("resize_reiserfs: block (%lu) has invalid format\n", block);
    }
	
    if (buffer_dirty(bh)) {
	mark_buffer_uptodate(bh,1);
	bwrite(bh);
    }
	
    brelse(bh);	

    new_blocknr = move_generic_block(s, block, bnd, h);
    if (new_blocknr) {
	if (node_is_internal)
	    int_moved_cnt++;
	else
	    leaf_moved_cnt++;
    }
	
    return new_blocknr;
}

int shrink_fs(struct super_block *s, unsigned long blocks)
{
	unsigned long n_root_block;
	unsigned int bmap_nr_new;

	/* warn about alpha version */
	{
		int c;

		printf(
			"You are running ALPHA version of reiserfs shrinker.\n"
			"This version is only for testing or VERY CAREFUL use.\n"
			"Backup of you data is recommended.\n\n"
			"Do you want to continue? [y/N]:"
			);
		c = getchar();
		if (c != 'y' && c != 'Y')
			exit(1);
	}

	bmap_nr_new = (blocks - 1) / (8 * s->s_blocksize) + 1;
	
	/* is shrinking possible ? */
	if (SB_BLOCK_COUNT(s) - blocks > SB_FREE_BLOCKS(s) + SB_BMAP_NR(s) - bmap_nr_new) {
	    fprintf(stderr, "resize_reiserfs: can\'t shrink fs; too many blocks already allocated\n");
	    return -1;
	}
	
	/* calculate number of data blocks */		
	blocks_used = 
	    SB_BLOCK_COUNT(s)
	    - SB_FREE_BLOCKS(s)
	    - SB_BMAP_NR(s)
	    - get_journal_size(s)
	    - REISERFS_DISK_OFFSET_IN_BYTES / s->s_blocksize
	    - 2; /* superblock itself and 1 descriptor after the journal */

	bmp = create_bitmap_from_sb(s);
	if (!bmp) 
		die ("resize_reiserfs: read bitmap failed\n");

	unused_block = 1;

	if (opt_verbose) {
		printf("Processing the tree: ");
		fflush(stdout);
	}

	n_root_block = move_formatted_block(s, SB_ROOT_BLOCK(s), blocks, 0);
	if (n_root_block) {
		PUT_SB_ROOT_BLOCK(s, n_root_block);
	}

	if (opt_verbose)
	    printf ("\n\nnodes processed (moved):\n"
		    "int        %lu (%lu),\n"
		    "leaves     %lu (%lu),\n" 
		    "unfm       %lu (%lu),\n"
		    "total      %lu (%lu).\n\n",
		    int_node_cnt, int_moved_cnt,
		    leaf_node_cnt, leaf_moved_cnt, 
		    unfm_node_cnt, unfm_moved_cnt,
		    (unsigned long)total_node_cnt, total_moved_cnt);

	if (block_count_mismatch) {
	    fprintf(stderr, "resize_reiserfs: data block count %lu"
		    " doesn\'t match data block count %lu from super block\n",
		    (unsigned long)total_node_cnt, blocks_used);
	}
#if 0
	printf("check for used blocks in truncated region\n");
	{
		unsigned long l;
		for (l = blocks; l < SB_BLOCK_COUNT(s); l++)
			if (is_block_used(bmp, l))
				printf("<%lu>", l);
		printf("\n");
	}
#endif /* 0 */
	
	PUT_SB_FREE_BLOCKS(s, SB_FREE_BLOCKS(s) - (SB_BLOCK_COUNT(s) - blocks) + (SB_BMAP_NR(s) - bmap_nr_new));
	PUT_SB_BLOCK_COUNT(s, blocks);
	PUT_SB_BMAP_NR(s, bmap_nr_new);
	
	truncate_bitmap(bmp, blocks);
	sync_bitmap(bmp);
	free_bitmap(bmp);

	return 0;
}
